# Generated by Django 5.1.3 on 2026-02-05 15:21

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("uk_geo_utils", "0013_rename_bua24_onspd_bua24cd_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="onspd",
            name="bua24cd",
            field=models.CharField(
                blank=True, help_text="Built-up Area (BUA)", max_length=9
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="cal24cd",
            field=models.CharField(
                blank=True, help_text="Cancer Alliance (CAL)", max_length=9
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="educ23cd",
            field=models.CharField(
                blank=True,
                help_text="Local Learning and Skills Council (LLSC)/ Dept. of Children, Education, Lifelong Learning and Skills (DCELLS)/ Enterprise Region (ER)",
                max_length=9,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="icb23cd",
            field=models.CharField(
                blank=True,
                help_text="Integrated Care Board (ICB)",
                max_length=9,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="imd20ind",
            field=models.CharField(
                blank=True,
                help_text="Index of Multiple Deprivation (IMD)for 2011 LSOAs",
                max_length=5,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="itl25cd",
            field=models.CharField(
                blank=True,
                help_text="International Territorial Level (former NUTS)",
                max_length=10,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="lat",
            field=models.CharField(
                blank=True, help_text="Decimal degrees latitude", max_length=10
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="lep21cd1",
            field=models.CharField(
                blank=True,
                help_text="Local Enterprise Partnership (LEP) - first instance",
                max_length=9,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="lep21cd2",
            field=models.CharField(
                blank=True,
                help_text="Local Enterprise Partnership (LEP) - second instance",
                max_length=9,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="long",
            field=models.CharField(
                blank=True, help_text="Decimal degrees longitude", max_length=10
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="lsoa01cd",
            field=models.CharField(
                blank=True,
                help_text="2001 Census Lower Layer Super Output Area (LSOA)/ Data Zone (DZ)/ Super Output Area (SOA)",
                max_length=9,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="lsoa11cd",
            field=models.CharField(
                blank=True,
                help_text="2011 Census Lower Layer Super Output Area (LSOA)/ Data Zone (DZ)/ SOA",
                max_length=9,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="lsoa21cd",
            field=models.CharField(
                blank=True,
                help_text="2021 Census Lower Layer Super Output Area (LSOA)/Super Data Zone (SDZ)",
                max_length=9,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="msoa01cd",
            field=models.CharField(
                blank=True,
                help_text="2001 Census Middle Layer Super Output Area (MSOA)/ Intermediate Zone (IZ)",
                max_length=9,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="msoa11cd",
            field=models.CharField(
                blank=True,
                help_text="2011 Census Middle Layer Super Output Area (MSOA)/Intermediate Zone (IZ)",
                max_length=9,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="msoa21cd",
            field=models.CharField(
                blank=True,
                help_text="2021 Census Middle Layer Super Output Area (MSOA)",
                max_length=9,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="npark16cd",
            field=models.CharField(
                blank=True, help_text="National park", max_length=9
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="oa01cd",
            field=models.CharField(
                blank=True,
                help_text="2001 Census Output Area (OA)",
                max_length=10,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="oa11cd",
            field=models.CharField(
                blank=True,
                help_text="2011 Census Output Area (OA)/Small Area (SA)",
                max_length=9,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="oa21cd",
            field=models.CharField(
                blank=True,
                help_text="2021 Census Output Area (OA)/Data Zone (DZ)",
                max_length=9,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="oac01ind",
            field=models.CharField(
                blank=True,
                help_text=" 2001 Census Output Area classification (OAC)",
                max_length=3,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="oac11ind",
            field=models.CharField(
                blank=True,
                help_text="2011 Census Output Area classification (OAC)",
                max_length=3,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="pco19cd",
            field=models.CharField(
                blank=True,
                help_text="Primary Care Organisatioin (PCO): Care Trust (PCT)/ Care Trust/ Care Trust Plus (CT)/ Local Health Board (LHB)/ Community Health Partnership (CHP)/ Local Commissioning Group (LCG)/ Primary Healthcare Directorate (PHD)",
                max_length=9,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="pfa23cd",
            field=models.CharField(
                blank=True, help_text="Police Force Area (PFA)", max_length=9
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="ruc01ind",
            field=models.CharField(
                blank=True,
                help_text="2001 Census urban/rural indicator",
                max_length=1,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="ruc11ind",
            field=models.CharField(
                blank=True,
                help_text="2011 Census rural-urban classification",
                max_length=2,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="ruc21ind",
            field=models.CharField(
                blank=True,
                help_text="2021 Census rural-urban classification",
                max_length=4,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="sicbl24cd",
            field=models.CharField(
                blank=True,
                help_text="Sub ICB Location (LOC)/ Local Health Board (LHB)/ Community Health Partnership (CHP)/ Local Commissioning Group (LCG)/ Primary Healthcare Directorate (PHD)",
                max_length=9,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="ssr95cd",
            field=models.CharField(
                blank=True,
                help_text="Standard Statistical Region (SSR)",
                max_length=9,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="wdcas03cd",
            field=models.CharField(
                blank=True,
                help_text="Census Area Statistics (CAS) ward",
                max_length=6,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="wdstl05cd",
            field=models.CharField(
                blank=True,
                help_text="2005 ‘statistical’ ward (England and Wales only)",
                max_length=6,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="wz11cd",
            field=models.CharField(
                blank=True,
                help_text="2011 Census Workplace Zone (WZ)",
                max_length=9,
            ),
        ),
    ]



# Generated by Django 5.1.3 on 2026-02-05 15:19

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        (
            "uk_geo_utils",
            "0012_rename_ced_onspd_ced25cd_rename_ctry_onspd_ctry25cd_and_more",
        ),
    ]

    operations = [
        migrations.RenameField(
            model_name="onspd",
            old_name="bua24",
            new_name="bua24cd",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="calncv",
            new_name="cal24cd",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="teclec",
            new_name="educ23cd",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="icb",
            new_name="icb23cd",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="imd",
            new_name="imd20ind",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="itl",
            new_name="itl25cd",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="lep1",
            new_name="lep21cd1",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="lep2",
            new_name="lep21cd2",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="lsoa01",
            new_name="lsoa01cd",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="lsoa11",
            new_name="lsoa11cd",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="lsoa21",
            new_name="lsoa21cd",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="msoa01",
            new_name="msoa01cd",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="msoa11",
            new_name="msoa11cd",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="msoa21",
            new_name="msoa21cd",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="npark",
            new_name="npark16cd",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="oa01",
            new_name="oa01cd",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="oa11",
            new_name="oa11cd",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="oa21",
            new_name="oa21cd",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="oac01",
            new_name="oac01ind",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="oac11",
            new_name="oac11ind",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="pct",
            new_name="pco19cd",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="pfa",
            new_name="pfa23cd",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="ur01ind",
            new_name="ruc01ind",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="ru11ind",
            new_name="ruc11ind",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="sicbl",
            new_name="sicbl24cd",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="casward",
            new_name="wdcas03cd",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="statsward",
            new_name="wdstl05cd",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="wz11",
            new_name="wz11cd",
        ),
        migrations.RemoveField(
            model_name="onspd",
            name="streg",
        ),
        migrations.AddField(
            model_name="onspd",
            name="ruc21ind",
            field=models.CharField(blank=True, max_length=4),
        ),
        migrations.AddField(
            model_name="onspd",
            name="ssr95cd",
            field=models.CharField(blank=True, max_length=9),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="ced25cd",
            field=models.CharField(
                blank=True, help_text="County Electoral Division", max_length=9
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="ctry25cd",
            field=models.CharField(
                blank=True, help_text="Country", max_length=9
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="cty25cd",
            field=models.CharField(
                blank=True, help_text="County", max_length=9
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="dointr",
            field=models.CharField(
                blank=True, help_text="Date of introduction", max_length=6
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="doterm",
            field=models.CharField(
                blank=True, help_text="Date of termination", max_length=6
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="east1m",
            field=models.CharField(
                blank=True,
                help_text="National grid reference - Easting",
                max_length=6,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="eer20cd",
            field=models.CharField(
                blank=True, help_text="European Electoral Region", max_length=9
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="gridind",
            field=models.CharField(
                blank=True,
                help_text="Grid reference positional quality indicator",
                max_length=1,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="hlth19cd",
            field=models.CharField(
                blank=True,
                help_text="Former Strategic Health Authority (SHA)/ Local Health Board (LHB)/ Health Board (HB)/ Health Authority (HA)/ Health & Social Care Board (HSCB)",
                max_length=9,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="lad25cd",
            field=models.CharField(
                blank=True,
                help_text="Local authority; district (LAD): unitary authority (UA)/non- metroploitan district (NMD)/ metropolitan district (MD)/ London borough (LB)/ council area (CA)/district council area (DCA)",
                max_length=9,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="nhser24cd",
            field=models.CharField(
                blank=True,
                help_text="NHS England (Region) (NHS ER)",
                max_length=9,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="north1m",
            field=models.CharField(
                blank=True,
                help_text="National grid reference - Northing",
                max_length=7,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="parncp25cd",
            field=models.CharField(
                blank=True,
                help_text="Civil parish/non- civil parish/ community",
                max_length=9,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="pcd7",
            field=models.CharField(
                blank=True,
                help_text="Unit postcode – 7 character version",
                max_length=7,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="pcd8",
            field=models.CharField(
                blank=True,
                help_text="Unit postcode – 8 character version",
                max_length=8,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="pcds",
            field=models.CharField(
                blank=True,
                help_text="Unit postcode - variable length (e-Gif) version",
                max_length=8,
                primary_key=True,
                serialize=False,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="pcon24cd",
            field=models.CharField(
                blank=True,
                help_text="Westminster parliamentary constituency",
                max_length=9,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="rgn25cd",
            field=models.CharField(
                blank=True, help_text="Region (former GOR)", max_length=9
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="ttwa15cd",
            field=models.CharField(
                blank=True, help_text="Travel to Work Area (TTWA)", max_length=9
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="usrtypind",
            field=models.CharField(
                blank=True, help_text="Postcode user type", max_length=1
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="wd25cd",
            field=models.CharField(
                blank=True, help_text="(Electoral) ward/division", max_length=9
            ),
        ),
    ]



# Generated by Django 5.1 on 2025-02-26 15:19

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        (
            "uk_geo_utils",
            "0010_rename_bua11_onspd_bua22_rename_nuts_onspd_itl_and_more",
        ),
    ]

    operations = [
        migrations.RenameField(
            model_name="onspd",
            old_name="bua22",
            new_name="bua24",
        ),
    ]



# Generated by Django 5.1.3 on 2026-02-05 14:44

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("uk_geo_utils", "0011_rename_bua22_onspd_bua24"),
    ]

    operations = [
        migrations.RenameField(
            model_name="onspd",
            old_name="ced",
            new_name="ced25cd",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="ctry",
            new_name="ctry25cd",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="oscty",
            new_name="cty25cd",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="oseast1m",
            new_name="east1m",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="eer",
            new_name="eer20cd",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="osgrdind",
            new_name="gridind",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="oshlthau",
            new_name="hlth19cd",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="oslaua",
            new_name="lad25cd",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="nhser",
            new_name="nhser24cd",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="osnrth1m",
            new_name="north1m",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="parish",
            new_name="parncp25cd",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="pcd",
            new_name="pcd7",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="pcd2",
            new_name="pcd8",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="pcon",
            new_name="pcon24cd",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="rgn",
            new_name="rgn25cd",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="ttwa",
            new_name="ttwa15cd",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="usertype",
            new_name="usrtypind",
        ),
        migrations.RenameField(
            model_name="onspd",
            old_name="osward",
            new_name="wd25cd",
        ),
    ]



import abc
import shutil
import tempfile
import urllib.request
import zipfile
from pathlib import Path

import psutil
from django.core.management.base import BaseCommand
from django.db import DEFAULT_DB_ALIAS, connections, transaction


def unzip(filepath):
    zip_file = zipfile.ZipFile(filepath, "r")
    tmpdir = tempfile.mkdtemp()
    zip_file.extractall(tmpdir)
    return tmpdir


def check_memory(required_memory: int = 2):
    # Downloading, unzipping and working with the ONSPD
    # requires a decent chunk of memory to play with.
    # Running this import on a tiny instance like a
    # t2.micro will cause an Out Of Memory error

    # By default ensure we've got >2Gb total before we start
    mem = psutil.virtual_memory()
    gb = ((mem.total / 1024) / 1024) / 1024
    return gb >= required_memory


class BaseImporter(BaseCommand):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.foreign_key_constraints = None
        self.indexes = None
        self.primary_key_constraint = None
        self.tempdir = None
        self.data_path = None
        self.cursor = None
        self.table_name = self.get_table_name()
        self.temp_table_name = self.table_name + "_temp"

    def add_arguments(self, parser):
        group = parser.add_mutually_exclusive_group(required=True)
        group.add_argument("--url", action="store")
        group.add_argument("--data-path", action="store")
        parser.add_argument("--database", default=DEFAULT_DB_ALIAS)

    @abc.abstractmethod
    def get_table_name(self) -> str:
        pass

    def get_data_path(self, options):
        data_path = None

        if options.get("data_path"):
            self.data_path = options["data_path"]

        if url := options.get("url"):
            self.stdout.write(f"Downloading data from {url}")
            tmp = tempfile.NamedTemporaryFile()
            urllib.request.urlretrieve(url, tmp.name)
            self.tempdir = unzip(tmp.name)
            self.data_path = Path(self.tempdir) / "Data"

        return data_path

    @abc.abstractmethod
    def import_data_to_temp_table(self):
        pass

    def get_index_statements(self):
        self.cursor.execute(f"""
            SELECT tablename, indexname, indexdef 
            FROM pg_indexes 
            WHERE tablename='{self.table_name}' 
        """)
        results = self.cursor.fetchall()

        indexes = []
        for row in results:
            original_index_name = row[1]
            original_index_create_statement = row[2]
            temp_index_name = original_index_name + "_temp"
            temp_index_create_statement = self.make_temp_index_create_statement(
                original_index_create_statement,
                original_index_name,
                temp_index_name,
            )
            rename_temp_index_statement = (
                f"ALTER INDEX {temp_index_name} RENAME TO {original_index_name}"
            )
            indexes.append(
                {
                    "index_name": original_index_name,
                    "temp_index_name": temp_index_name,
                    "original_index_create_statement": original_index_create_statement,
                    "temp_index_create_statement": temp_index_create_statement,
                    "rename_temp_index_statement": rename_temp_index_statement,
                }
            )

        return indexes

    def make_temp_index_create_statement(
        self,
        original_index_create_statement,
        original_index_name,
        temp_index_name,
    ):
        # we expect the statement to be of the form
        # CREATE [UNIQUE] INDEX $index ON $table USING $fields"
        temp_index_create_statement = original_index_create_statement.replace(
            f"INDEX {original_index_name}",
            f"INDEX IF NOT EXISTS {temp_index_name}",
        )
        return temp_index_create_statement.replace(
            f"ON public.{self.table_name}", f"ON public.{self.temp_table_name}"
        )

    def build_temp_indexes(self):
        self.stdout.write(f"Building indexes on {self.temp_table_name}...")
        for index in self.indexes:
            self.stdout.write(
                f"Executing: {index['temp_index_create_statement']}"
            )
            self.cursor.execute(index["temp_index_create_statement"])

    def get_primary_key_constraint(self):
        pkey_sql = f"""
            SELECT conname, pg_get_constraintdef(oid) 
            FROM pg_constraint 
            WHERE conrelid = '{self.table_name}'::regclass AND contype = 'p';
        """
        self.cursor.execute(pkey_sql)
        results = self.cursor.fetchall()
        num_keys = len(results)
        if num_keys != 1:
            raise ValueError(
                f"Expected there to be 1 primary key. But {num_keys} found."
            )

        self.stdout.write("Found primary key constraint")
        constraint_name = results[0][0]
        temp_name = constraint_name + "_temp"
        constraintdef = results[0][1]
        return {
            "constraint_name": constraint_name,
            "temp_name": temp_name,
            "constraintdef": constraintdef,
            "temp_constraint_create_statement": f"ALTER TABLE {self.temp_table_name} ADD CONSTRAINT {temp_name} {constraintdef}",
        }

    def add_temp_primary_key(self):
        self.stdout.write(f"Adding primary key to {self.temp_table_name}...")
        self.stdout.write(
            f"Executing: {self.primary_key_constraint['temp_constraint_create_statement']}"
        )
        self.cursor.execute(
            self.primary_key_constraint["temp_constraint_create_statement"]
        )

    def get_foreign_key_constraints(self):
        fkey_sql = f"""
            SELECT conname AS constraint_name, confrelid::regclass::text AS refrenced_table, pg_get_constraintdef(oid), conrelid::regclass::text AS referencing_table
            FROM pg_constraint 
            WHERE contype = 'f'
                AND (
                    conrelid = '{self.table_name}'::regclass 
                    OR
                    confrelid = '{self.table_name}'::regclass
                )
        """
        self.cursor.execute(fkey_sql)
        results = self.cursor.fetchall()

        self.stdout.write(
            f"Found {len(results)} foreign key constraints, where {self.table_name} is the referencing or referenced table"
        )

        fk_constraints = []
        seen = set()

        for row in results:
            constraint_name = row[0]
            if constraint_name not in seen:
                seen.add(constraint_name)
                # referenced_table_name = row[1]
                constraintdef = row[2]
                referencing_table = row[3]

                fk_constraints.append(
                    {
                        "constraint_name": constraint_name,
                        "create_statement": f"ALTER TABLE {referencing_table} ADD CONSTRAINT {constraint_name} {constraintdef}",
                        "delete_statement": f"ALTER TABLE {referencing_table} DROP CONSTRAINT IF EXISTS {constraint_name}",
                    }
                )

        return fk_constraints

    def drop_foreign_keys(self):
        self.stdout.write("Dropping foreign keys...")
        for constraint in self.foreign_key_constraints:
            self.stdout.write(f"Executing: {constraint['delete_statement']}")
            self.cursor.execute(constraint["delete_statement"])

    def check_foreign_key_exists(self, constraint_name):
        # This check is especially useful when we're importing addressbase and uprntocouncil in WDIV.
        # This is because there is a fk from one to the other, so they both pick up the same fk.
        # You can't do a try/except on the create because then any subsequent commands will fail with
        # SQLSTATE code:
        #  > The 25P02 error code in PostgreSQL is associated with the “in_failed_sql_transaction” state.
        #  > This error indicates that you are trying to execute an SQL command after a previous command
        #  > in the same transaction has failed. Once a transaction encounters an error, it becomes ‘tainted’,
        #  > and PostgreSQL will not allow any further SQL commands to be executed until the transaction is
        #  > either rolled back or the failed command is resolved with a savepoint.
        # https://philipmcclarence.com/how-to-diagnose-and-fix-the-25p02-in_failed_sql_transaction-error-code-in-postgres/

        self.cursor.execute(f"""
            SELECT 1
            FROM pg_constraint 
            WHERE contype = 'f'
                AND conname = '{constraint_name}'
        """)
        return bool(self.cursor.fetchone())

    def add_foreign_keys(self):
        self.stdout.write("Creating foreign keys...")
        for constraint in self.foreign_key_constraints:
            if self.check_foreign_key_exists(constraint["constraint_name"]):
                self.stdout.write(
                    f"Foreign key {constraint['constraint_name']} already exists - skipping"
                )
                continue

            self.stdout.write(f"Executing: {constraint['create_statement']}")
            self.cursor.execute(constraint["create_statement"])

    def create_temp_table(self):
        self.stdout.write(
            f"Creating temp table called {self.temp_table_name}..."
        )
        self.cursor.execute(f"DROP TABLE IF EXISTS {self.temp_table_name};")
        create_statement = f"CREATE TABLE {self.temp_table_name} AS SELECT * FROM {self.table_name} LIMIT 0;"
        self.stdout.write(f"Executing: {create_statement}")
        self.cursor.execute(create_statement)

    def alter_temp_table_replica_identity(self, identity):
        alter_table_statment = (
            f"ALTER TABLE {self.temp_table_name} REPLICA IDENTITY {identity};"
        )
        self.stdout.write(f"Executing: {alter_table_statment}")
        self.cursor.execute(alter_table_statment)

    def drop_old_table(self):
        self.stdout.write("Dropping old table...")
        drop_table_statement = f"DROP TABLE {self.table_name} CASCADE "
        self.stdout.write(f"Executing: {drop_table_statement}")
        self.cursor.execute(drop_table_statement)

    def rename_temp_table(self):
        self.stdout.write("Renaming temp table...")
        rename_table_statement = (
            f"ALTER TABLE {self.temp_table_name} RENAME TO {self.table_name}"
        )
        self.stdout.write(f"Executing: {rename_table_statement}")
        self.cursor.execute(rename_table_statement)
        self.stdout.write("Renaming primary key...")
        primary_key_rename_statement = f"ALTER TABLE {self.table_name} RENAME CONSTRAINT {self.primary_key_constraint['temp_name']} TO {self.primary_key_constraint['constraint_name']}"
        self.stdout.write(f"Executing: {primary_key_rename_statement}")
        self.cursor.execute(primary_key_rename_statement)
        self.rename_temp_indexes()

    def rename_temp_indexes(self):
        index_rename_statements = []
        for index in self.indexes:
            if (
                index["index_name"]
                == self.primary_key_constraint["constraint_name"]
            ):
                self.stdout.write(
                    f"Skipping rename of {index['index_name']} because renaming primary key constraint renamed it already"
                )
            else:
                index_rename_statements.append(
                    index["rename_temp_index_statement"]
                )

        for statement in index_rename_statements:
            self.stdout.write(f"Executing: {statement}")
            self.cursor.execute(statement)

    def check_for_other_constraints(self):
        self.stdout.write(
            "Checking for non primary key/foreign key constraints..."
        )
        self.cursor.execute(f"""
            SELECT 1
            FROM pg_constraint
            WHERE
                contype NOT IN ('p', 'f')
                AND
                conrelid = '{self.table_name}'::regclass;
        """)
        result = self.cursor.fetchone()
        if result:
            raise (
                Exception(
                    "Non primary key/foreign key constraints found. Aborting import, to avoid overwriting."
                )
            )
        self.stdout.write("...none found. Continuing.")

    def get_constraints_and_index_statements(self):
        self.stdout.write(
            f"Getting constraints and indexes for {self.table_name}"
        )
        self.primary_key_constraint = self.get_primary_key_constraint()
        self.indexes = self.get_index_statements()
        self.foreign_key_constraints = self.get_foreign_key_constraints()
        self.check_for_other_constraints()

    def handle(self, *args, **options):
        if not check_memory():
            raise Exception(
                "This instance has less than the recommended memory. Try running the import from a larger instance."
            )

        db_name = options["database"]
        self.connection = connections[db_name]
        self.cursor = self.connection.cursor()

        self.stdout.write(
            f"Connected to: {self.connection.settings_dict['NAME']} @ {self.connection.settings_dict['HOST'] if self.connection.settings_dict['HOST'] else 'localhost'}"
        )

        self.get_data_path(options)

        self.get_constraints_and_index_statements()

        try:
            # Create empty temp tables
            self.create_temp_table()

            # Set temp table replica identity to full
            self.alter_temp_table_replica_identity("FULL")

            # import data into the temp table
            self.import_data_to_temp_table()

            # Add temp primary keys
            self.add_temp_primary_key()

            # Add temp indexes
            self.build_temp_indexes()

            # Set temp table replica identity to default
            self.alter_temp_table_replica_identity("DEFAULT")

            with transaction.atomic(using=db_name):
                # Drop Foreign keys
                if self.foreign_key_constraints:
                    self.drop_foreign_keys()

                # drop old table
                self.drop_old_table()

                # Rename temp table to original names, pkey and indexes
                self.rename_temp_table()

                # Add Foreign keys
                if self.foreign_key_constraints:
                    self.add_foreign_keys()

        finally:
            self.db_cleanup()
            self.file_cleanup()
            self.report_on_replicaton_status()

        self.stdout.write("...done")

    def db_cleanup(self):
        self.stdout.write(
            f"Make sure {self.table_name} is set to replicate identity default"
        )
        alter_table_statment = (
            f"ALTER TABLE {self.table_name} REPLICA IDENTITY DEFAULT;"
        )
        self.stdout.write(f"Executing: {alter_table_statment}")
        self.cursor.execute(alter_table_statment)

        self.stdout.write("Dropping temp table if exists...")
        self.cursor.execute(
            f"DROP TABLE IF EXISTS {self.temp_table_name} CASCADE;"
        )

    def file_cleanup(self):
        if self.tempdir:
            self.stdout.write(f"Cleaning up temp files in {self.tempdir}")
            try:
                shutil.rmtree(self.tempdir)
            except OSError:
                self.stdout.write("Failed to clean up temp files.")

    def report_on_replicaton_status(self):
        if not self.connection.settings_dict["HOST"]:
            self.stdout.write(
                "Command appears to have been run against a local database. Skipping replication checks."
            )
            return
        dbname = self.cursor.db.connection.info.dbname
        query_statement = f"""
            SELECT slot_name, wal_status, active, conflicting 
            FROM pg_replication_slots
            WHERE database = '{dbname}';
        """
        self.cursor.execute(query_statement)
        results = self.cursor.fetchall()
        if not results:
            self.stdout.write(
                self.style.NOTICE(
                    "No replication slots found. If you think there should be some either the --database flag was wrong, or something has broken.",
                )
            )
            return

        columns = [desc[0] for desc in self.cursor.description]
        result_dicts = []
        for row in results:
            result_dicts.append(dict(zip(columns, row)))

        self.stdout.write(self.style.SQL_TABLE("Replication status:"))
        broken_slots = False
        for result in result_dicts:
            if result["active"] and not result["conflicting"]:
                self.stdout.write(
                    self.style.SUCCESS(
                        f"{result['slot_name']} is active and is not conflicting. 'wal_status' is {result['wal_status']}."
                    )
                )
            else:
                self.stdout.write(
                    self.style.WARNING(
                        f"{result['slot_name']} Doesn't seem healthy:"
                        f"\n    'active': {result['active']}."
                        f"\n    'wal_status': {result['wal_status']}."
                        f"\n    'conflicting': {result['conflicting']}."
                    )
                )
                broken_slots = True

        if broken_slots:
            self.stdout.write(
                self.style.NOTICE(
                    "Some replication slots are broken. You can probably fix this by cycling out the instances."
                )
            )



from django.contrib.gis.db import models

try:
    from django.contrib.gis.db.models.manager import GeoManager
except ImportError:
    from django.db.models import Manager as GeoManager


class CachedGetMixin:
    def get_cached(self, pk):
        for record in self:
            if record.pk == pk:
                return record
        raise self.model.DoesNotExist()


class AddressQuerySet(models.QuerySet, CachedGetMixin):
    @property
    def centroid(self):
        if not self:
            return None

        if len(self) == 1:
            return self[0].location

        base_point = self[0].location
        poly = base_point.union(self[1].location)
        for m in self:
            poly = poly.union(m.location)

        return poly.centroid


class AbstractAddressManager(GeoManager):
    def get_queryset(self):
        return AddressQuerySet(self.model, using=self._db)


class AbstractAddress(models.Model):
    uprn = models.CharField(primary_key=True, max_length=100)
    address = models.TextField(blank=True)
    postcode = models.CharField(blank=True, max_length=15, db_index=True)
    location = models.PointField(null=True, blank=True)
    addressbase_postal = models.CharField(blank=False, max_length=1)
    objects = AbstractAddressManager()

    class Meta:
        abstract = True


class Address(AbstractAddress):
    pass


class OnsudQuerySet(models.QuerySet, CachedGetMixin):
    pass


class AbstractOnsudManager(GeoManager):
    def get_queryset(self):
        return OnsudQuerySet(self.model, using=self._db)


class AbstractOnsud(models.Model):
    uprn = models.CharField(primary_key=True, max_length=12)
    cty = models.CharField(blank=True, max_length=9)
    ced = models.CharField(blank=True, max_length=9)
    lad = models.CharField(blank=True, max_length=9)
    ward = models.CharField(blank=True, max_length=9)
    parish = models.CharField(blank=True, max_length=9)
    hlthau = models.CharField(blank=True, max_length=9)
    ctry = models.CharField(blank=True, max_length=9)
    rgn = models.CharField(blank=True, max_length=9)
    pcon = models.CharField(blank=True, max_length=9)
    eer = models.CharField(blank=True, max_length=9)
    ttwa = models.CharField(blank=True, max_length=9)
    nuts = models.CharField(blank=True, max_length=9)
    park = models.CharField(blank=True, max_length=9)
    oa11 = models.CharField(blank=True, max_length=9)
    lsoa11 = models.CharField(blank=True, max_length=9)
    msoa11 = models.CharField(blank=True, max_length=9)
    wz11 = models.CharField(blank=True, max_length=9)
    ccg = models.CharField(blank=True, max_length=9)
    bua11 = models.CharField(blank=True, max_length=9)
    buasd11 = models.CharField(blank=True, max_length=9)
    ruc11 = models.CharField(blank=True, max_length=2)
    oac11 = models.CharField(blank=True, max_length=3)
    lep1 = models.CharField(blank=True, max_length=9)
    lep2 = models.CharField(blank=True, max_length=9)
    pfa = models.CharField(blank=True, max_length=9)
    imd = models.CharField(blank=True, max_length=5)
    objects = AbstractOnsudManager()

    def _get_oscty(self):
        return self.cty

    def _get_oslaua(self):
        return self.lad

    def _get_osward(self):
        return self.ward

    def _get_oshlthau(self):
        return self.hlthau

    def _get_ru11ind(self):
        return self.ruc11

    oscty = property(_get_oscty)
    oslaua = property(_get_oslaua)
    osward = property(_get_osward)
    oshlthau = property(_get_oshlthau)
    ru11ind = property(_get_ru11ind)

    class Meta:
        abstract = True


class Onsud(AbstractOnsud):
    pass


class AbstractOnspd(models.Model):
    pcd7 = models.CharField(
        blank=True,
        max_length=7,
        help_text="Unit postcode – 7 character version",
    )
    pcd8 = models.CharField(
        blank=True,
        max_length=8,
        help_text="Unit postcode – 8 character version",
    )
    pcds = models.CharField(
        blank=True,
        max_length=8,
        primary_key=True,
        help_text="Unit postcode - variable length (e-Gif) version",
    )
    dointr = models.CharField(
        blank=True, max_length=6, help_text="Date of introduction"
    )
    doterm = models.CharField(
        blank=True, max_length=6, help_text="Date of termination"
    )
    cty25cd = models.CharField(blank=True, max_length=9, help_text="County")
    ced25cd = models.CharField(
        blank=True, max_length=9, help_text="County Electoral Division"
    )
    lad25cd = models.CharField(
        blank=True,
        max_length=9,
        help_text="Local authority district (LAD): unitary authority (UA)/non- metroploitan district (NMD)/metropolitan district (MD)/London borough (LB)/council area (CA)/district council area (DCA)",
    )
    wd25cd = models.CharField(
        blank=True, max_length=9, help_text="(Electoral) ward/division"
    )
    parncp25cd = models.CharField(
        blank=True,
        max_length=9,
        help_text="Civil parish/non-civil parish/community",
    )
    usrtypind = models.CharField(
        blank=True, max_length=1, help_text="Postcode user type"
    )
    east1m = models.CharField(
        blank=True, max_length=6, help_text="National grid reference - Easting"
    )
    north1m = models.CharField(
        blank=True, max_length=7, help_text="National grid reference - Northing"
    )
    gridind = models.CharField(
        blank=True,
        max_length=1,
        help_text="Grid reference positional quality indicator",
    )
    hlth19cd = models.CharField(
        blank=True,
        max_length=9,
        help_text="Former Strategic Health Authority (SHA)/Local Health Board (LHB)/Health Board (HB)/Health Authority (HA)/Health & Social Care Board (HSCB)",
    )
    nhser24cd = models.CharField(
        blank=True, max_length=9, help_text="NHS England (Region) (NHS ER)"
    )
    ctry25cd = models.CharField(blank=True, max_length=9, help_text="Country")
    rgn25cd = models.CharField(
        blank=True, max_length=9, help_text="Region (former GOR)"
    )
    ssr95cd = models.CharField(
        blank=True, max_length=9, help_text="Standard Statistical Region (SSR)"
    )
    pcon24cd = models.CharField(
        blank=True,
        max_length=9,
        help_text="Westminster parliamentary constituency",
    )
    eer20cd = models.CharField(
        blank=True, max_length=9, help_text="European Electoral Region"
    )
    educ23cd = models.CharField(
        blank=True,
        max_length=9,
        help_text="Local Learning and Skills Council (LLSC)/Dept. of Children, Education, Lifelong Learning and Skills (DCELLS)/Enterprise Region (ER)",
    )
    ttwa15cd = models.CharField(
        blank=True, max_length=9, help_text="Travel to Work Area (TTWA)"
    )
    pco19cd = models.CharField(
        blank=True,
        max_length=9,
        help_text="Primary Care Organisatioin (PCO): Care Trust (PCT)/Care Trust/Care Trust Plus (CT)/Local Health Board (LHB)/Community Health Partnership (CHP)/Local Commissioning Group (LCG)/Primary Healthcare Directorate (PHD)",
    )
    itl25cd = models.CharField(
        blank=True,
        max_length=10,
        help_text="International Territorial Level (former NUTS)",
    )
    wdstl05cd = models.CharField(
        blank=True,
        max_length=6,
        help_text="2005 ‘statistical’ ward (England and Wales only)",
    )
    oa01cd = models.CharField(
        blank=True, max_length=10, help_text="2001 Census Output Area (OA)"
    )
    wdcas03cd = models.CharField(
        blank=True, max_length=6, help_text="Census Area Statistics (CAS) ward"
    )
    npark16cd = models.CharField(
        blank=True, max_length=9, help_text="National park"
    )
    lsoa01cd = models.CharField(
        blank=True,
        max_length=9,
        help_text="2001 Census Lower Layer Super Output Area (LSOA)/Data Zone (DZ)/Super Output Area (SOA)",
    )
    msoa01cd = models.CharField(
        blank=True,
        max_length=9,
        help_text="2001 Census Middle Layer Super Output Area (MSOA)/Intermediate Zone (IZ)",
    )
    ruc01ind = models.CharField(
        blank=True, max_length=1, help_text="2001 Census urban/rural indicator"
    )
    oac01ind = models.CharField(
        blank=True,
        max_length=3,
        help_text=" 2001 Census Output Area classification (OAC)",
    )
    oa11cd = models.CharField(
        blank=True,
        max_length=9,
        help_text="2011 Census Output Area (OA)/Small Area (SA)",
    )
    lsoa11cd = models.CharField(
        blank=True,
        max_length=9,
        help_text="2011 Census Lower Layer Super Output Area (LSOA)/Data Zone (DZ)/SOA",
    )
    msoa11cd = models.CharField(
        blank=True,
        max_length=9,
        help_text="2011 Census Middle Layer Super Output Area (MSOA)/Intermediate Zone (IZ)",
    )
    wz11cd = models.CharField(
        blank=True, max_length=9, help_text="2011 Census Workplace Zone (WZ)"
    )
    bua24cd = models.CharField(
        blank=True, max_length=9, help_text="Built-up Area (BUA)"
    )
    ruc11ind = models.CharField(
        blank=True,
        max_length=2,
        help_text="2011 Census rural-urban classification",
    )
    oac11ind = models.CharField(
        blank=True,
        max_length=3,
        help_text="2011 Census Output Area classification (OAC)",
    )
    lat = models.CharField(
        blank=True, max_length=10, help_text="Decimal degrees latitude"
    )
    long = models.CharField(
        blank=True, max_length=10, help_text="Decimal degrees longitude"
    )
    lep21cd1 = models.CharField(
        blank=True,
        max_length=9,
        help_text="Local Enterprise Partnership (LEP) - first instance",
    )
    lep21cd2 = models.CharField(
        blank=True,
        max_length=9,
        help_text="Local Enterprise Partnership (LEP) - second instance",
    )
    pfa23cd = models.CharField(
        blank=True, max_length=9, help_text="Police Force Area (PFA)"
    )
    imd20ind = models.CharField(
        blank=True,
        max_length=5,
        help_text="Index of Multiple Deprivation (IMD)for 2011 LSOAs",
    )
    cal24cd = models.CharField(
        blank=True, max_length=9, help_text="Cancer Alliance (CAL)"
    )
    oa21cd = models.CharField(
        blank=True,
        max_length=9,
        help_text="2021 Census Output Area (OA)/Data Zone (DZ)",
    )
    lsoa21cd = models.CharField(
        blank=True,
        max_length=9,
        help_text="2021 Census Lower Layer Super Output Area (LSOA)/Super Data Zone (SDZ)",
    )
    msoa21cd = models.CharField(
        blank=True,
        max_length=9,
        help_text="2021 Census Middle Layer Super Output Area (MSOA)",
    )
    ruc21ind = models.CharField(
        blank=True,
        max_length=4,
        help_text="2021 Census rural-urban classification",
    )
    icb23cd = models.CharField(
        blank=True, max_length=9, help_text="Integrated Care Board (ICB)"
    )
    sicbl24cd = models.CharField(
        blank=True,
        max_length=9,
        help_text="Sub ICB Location (LOC)/Local Health Board (LHB)/Community Health Partnership (CHP)/Local Commissioning Group (LCG)/Primary Healthcare Directorate (PHD)",
    )

    location = models.PointField(null=True, blank=True)
    objects = GeoManager()

    def _get_cty(self):
        """
        Note: these are not _exactly_ the same because

        'CTY' in ONSUD is:
        E10000002 - E10000034 = England (county);
        E11000001 - E11000007 = England (metropolitan county);
        E13000001 - E13000002 = England (Inner/Outer London);
        E99999999 (pseudo) = England (UA/MD/LB);
        W99999999 (pseudo) = Wales;
        S99999999 (pseudo) = Scotland

        whereas 'OSCTY' in ONSPD is
        E10000002 – E10000034 = England;
        E99999999 (pseudo) = England (UA/MD/LB);
        W99999999 (pseudo) = Wales;
        S99999999 (pseudo) = Scotland;
        """
        return self.cty25cd

    def _get_lad(self):
        return self.lad25cd

    def _get_ward(self):
        return self.wd25cd

    def _get_hlthau(self):
        return self.hlth19cd

    def _get_ruc11(self):
        return self.ruc11ind

    def _get_usertype(self):
        return self.usrtypind

    cty = property(_get_cty)
    lad = property(_get_lad)
    ward = property(_get_ward)
    hlthau = property(_get_hlthau)
    ruc11 = property(_get_ruc11)
    usertype = property(_get_usertype)

    class Meta:
        abstract = True


class Onspd(AbstractOnspd):
    pass



# Generated by Django 5.1.3 on 2026-03-04 15:32

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        (
            "uk_geo_utils",
            "0014_alter_onspd_bua24cd_alter_onspd_cal24cd_and_more",
        ),
    ]

    operations = [
        migrations.AlterField(
            model_name="onspd",
            name="educ23cd",
            field=models.CharField(
                blank=True,
                help_text="Local Learning and Skills Council (LLSC)/Dept. of Children, Education, Lifelong Learning and Skills (DCELLS)/Enterprise Region (ER)",
                max_length=9,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="hlth19cd",
            field=models.CharField(
                blank=True,
                help_text="Former Strategic Health Authority (SHA)/Local Health Board (LHB)/Health Board (HB)/Health Authority (HA)/Health & Social Care Board (HSCB)",
                max_length=9,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="lad25cd",
            field=models.CharField(
                blank=True,
                help_text="Local authority district (LAD): unitary authority (UA)/non- metroploitan district (NMD)/metropolitan district (MD)/London borough (LB)/council area (CA)/district council area (DCA)",
                max_length=9,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="lsoa01cd",
            field=models.CharField(
                blank=True,
                help_text="2001 Census Lower Layer Super Output Area (LSOA)/Data Zone (DZ)/Super Output Area (SOA)",
                max_length=9,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="lsoa11cd",
            field=models.CharField(
                blank=True,
                help_text="2011 Census Lower Layer Super Output Area (LSOA)/Data Zone (DZ)/SOA",
                max_length=9,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="msoa01cd",
            field=models.CharField(
                blank=True,
                help_text="2001 Census Middle Layer Super Output Area (MSOA)/Intermediate Zone (IZ)",
                max_length=9,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="parncp25cd",
            field=models.CharField(
                blank=True,
                help_text="Civil parish/non-civil parish/community",
                max_length=9,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="pco19cd",
            field=models.CharField(
                blank=True,
                help_text="Primary Care Organisatioin (PCO): Care Trust (PCT)/Care Trust/Care Trust Plus (CT)/Local Health Board (LHB)/Community Health Partnership (CHP)/Local Commissioning Group (LCG)/Primary Healthcare Directorate (PHD)",
                max_length=9,
            ),
        ),
        migrations.AlterField(
            model_name="onspd",
            name="sicbl24cd",
            field=models.CharField(
                blank=True,
                help_text="Sub ICB Location (LOC)/Local Health Board (LHB)/Community Health Partnership (CHP)/Local Commissioning Group (LCG)/Primary Healthcare Directorate (PHD)",
                max_length=9,
            ),
        ),
    ]
